# MIT License

# Copyright (c) 2025 Ashlin Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import argparse
import os
import re


def get_gbd_argparser():
    parser = argparse.ArgumentParser(description="GBD Benchmark Database")
    parser.add_argument("-d", "--db", type=gbd_db_type, default=os.environ.get("GBD_DB"), help="Specify database to work with")
    parser.add_argument("-v", "--verbose", action="store_true", help="Print additional (or diagnostic) information to stderr")
    return parser


def add_query_and_hashes_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("query", help="GBD Query", nargs="?")
    parser.add_argument(
        "--hashes", help="Explicitly select instances: Hashes can be passed as arguments to this option, but also via <stdin>.", nargs="*", default=[]
    )


def add_resource_limits_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("-j", "--jobs", default=1, type=int, help="Set number of parallel jobs")
    parser.add_argument(
        "-t",
        "--tlim",
        default=5000,
        type=int,
        help="Time limit (sec) per instance for 'init' sub-commands (also used for score calculation in 'eval' and 'plot')",
    )
    parser.add_argument("-m", "--mlim", default=2000, type=int, help="Memory limit (MB) per instance for 'init' sub-commands")
    parser.add_argument("-f", "--flim", default=1000, type=int, help="File size limit (MB) per instance for 'init' sub-commands which create files")


### Argument Types for Input Sanitation in ArgParse Library
def directory_type(path):
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError("{0} is not a directory".format(path))
    if os.access(path, os.R_OK):
        return os.path.abspath(path)
    else:
        raise argparse.ArgumentTypeError("{0} is not readable".format(path))


def file_type(path):
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError("{0} is not a regular file".format(path))
    if os.access(path, os.R_OK):
        return os.path.abspath(path)
    else:
        raise argparse.ArgumentTypeError("{0} is not readable".format(path))


def column_type(s):
    pat = re.compile(r"^[a-zA-Z][a-zA-Z0-9_]*$")
    if not pat.match(s):
        raise argparse.ArgumentTypeError('Column "{0}" does not match regular expression {1}'.format(s, pat.pattern))
    return s


def key_value_type(s):
    tup = s.split("=", 1)
    if len(tup) != 2:
        raise argparse.ArgumentTypeError("key-value type: {0} must be separated by exactly one = ".format(s))
    return (column_type(tup[0]), tup[1])


def gbd_db_type(dbstr):
    if not dbstr:
        default = os.environ.get("GBD_DB")
        if not default:
            raise argparse.ArgumentTypeError("Datasources Missing: Set GBD_DB environment variable (Find databases here: https://benchmark-database.de)")
        return default  # .split(':')
    return dbstr



#!/usr/bin/python3

# MIT License

# Copyright (c) 2025 Ashlin Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import os
import sys
import traceback
import polars as pl

from gbd_core.api import GBD, GBDException
from gbd_core.grammar import ParserException
from gbd_core import util, contexts, schema
from gbd_core.util_argparse import *
from gbd_init.feature_extractors import generic_extractors
from gbd_init.instance_transformers import generic_transformers


### Command-Line Interface Entry Points
def cli_hash(api: GBD, args):
    from gbd_core.contexts import identify

    print(identify(args.path))


def cli_init_local(api: GBD, args):
    from gbd_init.feature_extractors import init_local

    rlimits = {"jobs": args.jobs, "tlim": args.tlim, "mlim": args.mlim, "flim": args.flim}
    init_local(api, rlimits, args.path, args.target)


def cli_init_generic(api: GBD, args):
    from gbd_init.feature_extractors import init_features_generic

    rlimits = {"jobs": args.jobs, "tlim": args.tlim, "mlim": args.mlim, "flim": args.flim}
    context = api.database.dcontext(args.target)
    df = api.query(args.query, args.hashes, [context + ":local"], collapse="MIN", group_by=context + ":hash")
    init_features_generic(args.initfuncname, api, rlimits, df, args.target)


def cli_trans_generic(api: GBD, args):
    from gbd_init.instance_transformers import transform_instances_generic

    rlimits = {"jobs": args.jobs, "tlim": args.tlim, "mlim": args.mlim, "flim": args.flim}
    transform_instances_generic(args.transfuncname, api, rlimits, args.query, args.hashes, args.target, args.source, args.collapse)


def cli_create(api: GBD, args):
    api.create_feature(args.name, args.unique, args.target)


def cli_delete(api: GBD, args):
    if (args.hashes and len(args.hashes) or args.values and len(args.values)) and args.name:
        if args.force or util.confirm("Delete attributes of given hashes and/or values from '{}'?".format(args.name)):
            api.reset_values(args.name, args.values, args.hashes)
    elif args.hashes and not args.name:
        if args.force or util.confirm("Delete given hashes entirely?".format(args.name)):
            api.delete_hashes(args.hashes)
    elif args.force or util.confirm("Delete feature '{}' and all associated attributes?".format(args.name)):
        api.delete_feature(args.name)


def cli_cleanup(api: GBD, args):
    if args.hashes and len(args.hashes):
        if args.force or util.confirm("Delete attributes of given hashes from all features?"):
            api.delete_hashes(args.hashes, args.target)


def cli_rename(api: GBD, args):
    api.rename_feature(args.old_name, args.new_name)


def cli_copy(api: GBD, args):
    api.copy_feature(args.old_name, args.new_name, args.target, args.query, args.hashes)


def cli_get(api: GBD, args):
    df: pl.DataFrame = api.query(args.query, args.hashes, args.resolve, args.collapse, args.group_by, args.join_type)
    if args.header:
        print(args.delimiter.join(df.columns))
    for row in df.iter_rows(named=True):
        print(args.delimiter.join([str(row[col]) if row[col] is not None else "[None]" for col in df.columns]))


def cli_set(api: GBD, args):
    hashes = api.query(args.query, args.hashes)["hash"].to_list()
    if args.create:
        hashes = list(set(hashes + args.hashes))
    if len(hashes) > 0:
        api.set_values(args.assign[0], args.assign[1], hashes)


def cli_info(api: GBD, args):
    if args.contexts:
        print("# Available Contexts: " + ", ".join(contexts.contexts()))
        for context in contexts.contexts():
            print()
            print("## " + contexts.description(context))
            print(" - Context Prefix: " + context)
            print(" - File Extensions: " + ",".join(contexts.suffixes(context)))
    elif args.name is None:
        print("# Available Data Sources: " + ", ".join(api.get_databases()))
        for dbname in api.get_databases():
            if len(api.get_features(dbname)):
                print()
                print("## " + api.get_database_path(dbname))
                print(" - Name: " + dbname)
                feat = api.get_features(dbname)
                print(" - Features: " + " ".join(feat))
                if args.verbose:
                    for f in feat:
                        info = api.database.find(":".join([dbname, f]))
                        print(info)
    else:
        info = api.get_feature_info(args.name)
        for key in info:
            print("{}: {}".format(key, info[key]))


def cli_server(api: GBD, args):
    from gbd_server import server

    util.eprint("Starting GBD Server on port {}...".format(args.port))
    util.eprint(r"""
Warning: All files referenced in the configured databases are now accessible on the specified port.
If you do not trust the source of the databases, do not run the server.
""")
    server.serve(api, args.port, args.logdir)


### Define Command-Line Interface and Map Sub-Commands to Methods
def main():
    parser = get_gbd_argparser()

    subparsers = parser.add_subparsers(help="Available Commands:", required=True, dest="gbd command")

    # INITIALIZATION
    parser_init = subparsers.add_parser("init", help="Initialize Database")
    add_resource_limits_arguments(parser_init)
    parser_init.add_argument("--target", help="Target database for new features (default: first db in list); also determines target context", default=None)

    parser_init_subparsers = parser_init.add_subparsers(help="Select Initialization Procedure:", required=True, dest="init what?")

    # init local paths:
    parser_init_local = parser_init_subparsers.add_parser("local", help="Initialize Local Hash/Path Entries")
    parser_init_local.add_argument("path", type=directory_type, help="Path to benchmarks")
    parser_init_local.set_defaults(func=cli_init_local)

    # hooks for generic feature extractors:
    for key in generic_extractors.keys():
        gex = generic_extractors[key]
        parser_init_generic = parser_init_subparsers.add_parser(key, help=gex["description"])
        add_query_and_hashes_arguments(parser_init_generic)
        parser_init_generic.set_defaults(func=cli_init_generic, initfuncname=key)

    # TRANSFORMATION
    parser_trans = subparsers.add_parser("transform", help="Transform Benchmarks")
    add_resource_limits_arguments(parser_trans)
    parser_trans.add_argument("--source", help="Source context", default=contexts.default_context())
    parser_trans.add_argument("--target", help="Target database; determines target context (default: first db in list)", default=None)

    parser_trans_subparsers = parser_trans.add_subparsers(help="Select Transformation Procedure:", required=True, dest="transform how?")

    # hooks for generic instance transformers:
    for key in generic_transformers.keys():
        gex = generic_transformers[key]
        parser_trans_generic = parser_trans_subparsers.add_parser(key, help=gex["description"])
        add_query_and_hashes_arguments(parser_trans_generic)
        parser_trans_generic.set_defaults(func=cli_trans_generic, transfuncname=key)
        parser_trans_generic.add_argument(
            "-c",
            "--collapse",
            default="group_concat",
            choices=["group_concat", "min", "max", "avg", "count", "sum", "none"],
            help="Specify a function for the handling of multiple feature values",
        )

    # GBD HASH
    parser_hash = subparsers.add_parser("hash", help="Print hash for a single file")
    parser_hash.add_argument("path", type=file_type, help="Path to one benchmark")
    parser_hash.set_defaults(func=cli_hash)

    # GBD GET $QUERY
    parser_get = subparsers.add_parser("get", help="Get data by query (or hash-list via stdin)")
    add_query_and_hashes_arguments(parser_get)
    parser_get.add_argument("-r", "--resolve", help="List of feature names to resolve against", nargs="+", default=[])
    parser_get.add_argument(
        "-c",
        "--collapse",
        default="group_concat",
        choices=["group_concat", "min", "max", "avg", "count", "sum", "none"],
        help="Specify a function for the handling of multiple feature values",
    )
    parser_get.add_argument("-g", "--group_by", default=None, help="Group by the specified feature as the key, rather than by the primary key")
    parser_get.add_argument("--join-type", help="Join Type: treatment of missing values", choices=["INNER", "OUTER", "LEFT"], default="LEFT")
    parser_get.add_argument("-d", "--delimiter", default=" ", help="CSV delimiter to use in output")
    parser_get.add_argument("-H", "--header", action="store_true", help="Include header information in output")
    parser_get.set_defaults(func=cli_get)

    # GBD SET
    parser_set = subparsers.add_parser("set", help="Set specified attribute-value for query result")
    parser_set.add_argument("assign", type=key_value_type, help="key=value")
    parser_set.add_argument(
        "-c", "--create", help="Create given hashes if they do not exist yet (otherwise intersect with existing hashes)", action="store_true"
    )
    add_query_and_hashes_arguments(parser_set)
    parser_set.set_defaults(func=cli_set)

    # CREATE/DELETE/MODIFY FEATURES
    parser_create = subparsers.add_parser("create", help="Create a new feature")
    parser_create.add_argument("name", type=column_type, help="Name of feature")
    parser_create.add_argument("-u", "--unique", help="Unique constraint: specify default-value of feature")
    parser_create.add_argument("--target", help="Target database (default: first in list)", default=None)
    parser_create.set_defaults(func=cli_create)

    parser_delete = subparsers.add_parser(
        "delete", help="Delete all values assiociated with given hashes (via argument or stdin) or remove feature if no hashes are given"
    )
    parser_delete.add_argument("--hashes", help="Hashes for which to delete values", nargs="*", default=[])
    parser_delete.add_argument("--values", help="Values to delete", nargs="*", default=[])
    parser_delete.add_argument("name", type=column_type, help="Name of feature (default: all)", nargs="?")
    parser_delete.add_argument("-f", "--force", action="store_true", help="Do not ask for confirmation")
    parser_delete.set_defaults(func=cli_delete)

    parser_cleanup = subparsers.add_parser("cleanup", help="Delete given hashes from all features")
    parser_cleanup.add_argument("--hashes", help="Hashes for which to delete values", nargs="*", default=[])
    parser_cleanup.add_argument("-f", "--force", action="store_true", help="Do not ask for confirmation")
    parser_cleanup.add_argument("--target", help="Target database (default: first in list)", default=None)
    parser_cleanup.set_defaults(func=cli_cleanup)

    parser_rename = subparsers.add_parser("rename", help="Rename feature")
    parser_rename.add_argument("old_name", type=column_type, help="Old name of feature")
    parser_rename.add_argument("new_name", type=column_type, help="New name of feature")
    parser_rename.set_defaults(func=cli_rename)

    parser_copy = subparsers.add_parser("copy", help="Copy feature")
    add_query_and_hashes_arguments(parser_copy)
    parser_copy.add_argument("--target", help="Target database (default: first in list)", default=None)
    parser_copy.add_argument("old_name", type=column_type, help="Old name of feature")
    parser_copy.add_argument("new_name", type=column_type, help="New name of feature")
    parser_copy.set_defaults(func=cli_copy)

    # GET META INFO
    parser_info = subparsers.add_parser("info", help="Print info about available features")
    parser_info.add_argument("-c", "--contexts", action="store_true", help="Print available contexts")
    parser_info.add_argument("name", type=column_type, help="Print info about specified feature", nargs="?")
    parser_info.set_defaults(func=cli_info)

    # RUN SERVER
    parser_server = subparsers.add_parser("serve", help="Run GBD Server")
    parser_server.add_argument("-p", "--port", help="Specify port on which to listen", default=os.environ.get("GBD_PORT") or 5000, type=int)
    parser_server.add_argument("-l", "--logdir", help="Specify directory for logfiles", default=os.environ.get("GBD_LOGS") or "./")
    parser_server.set_defaults(func=cli_server)

    # PARSE ARGUMENTS
    args = parser.parse_args()
    try:
        if hasattr(args, "hashes") and not sys.stdin.isatty():
            if not args.hashes or len(args.hashes) == 0:
                args.hashes = util.read_hashes()  # read hashes from stdin
        if hasattr(args, "target") and args.target is None:
            args.target = schema.Schema.dbname_from_path(args.db.split(os.pathsep)[0])
        if args.db is None or len(args.db) == 0:
            util.eprint("No database specified. Use -d or set GBD_DB environment variable.")
            sys.exit(1)
        with GBD(args.db.split(os.pathsep), args.verbose) as api:
            args.func(api, args)
    except ModuleNotFoundError as e:
        util.eprint("Module '{}' not found. Please install it.".format(e.name))
        if e.name == "gbdc":
            util.eprint("Find installation instructions at https://github.com/Udopia/gbdc")
        sys.exit(1)
    except ParserException as e:
        util.eprint("Failed to parse query: " + args.query)
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)
    except pl.exceptions.DataOrientationWarning as e:
        util.eprint(traceback.format_exc())
    except Exception as e:
        util.eprint("{}: {}".format(type(e), str(e)))
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)


if __name__ == "__main__":
    main()



# MIT License

# Copyright (c) 2025 Ashlin Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import pandas as pd
import os
import glob
import warnings
import polars as pl

from gbd_core.contexts import suffixes, identify, get_context_by_suffix
from gbd_core.api import GBD, GBDException
from gbd_core.util import eprint, confirm
from gbd_init.initializer import Initializer, InitializerException

try:
    from gbdc import (
        extract_base_features,
        base_feature_names,
        extract_gate_features,
        gate_feature_names,
        isohash,
        wcnfisohash,
        wcnf_base_feature_names,
        extract_wcnf_base_features,
        opb_base_feature_names,
        extract_opb_base_features,
        checksani,
        checksani_feature_names,
    )
except ImportError:

    def extract_base_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def base_feature_names():
        return []

    def extract_gate_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def gate_feature_names():
        return []

    def isohash(path):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def extract_wcnf_base_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def wcnf_base_feature_names():
        return []

    def extract_opb_base_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def opb_base_feature_names():
        return []

    def checksani(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def checksani_feature_names():
        return []


## GBDHash
def compute_hash(hash, path, limits):
    eprint("Hashing {}".format(path))
    hash = identify(path)
    return [("local", hash, path), ("filename", hash, os.path.basename(path))]


## ISOHash
def compute_isohash(hash, path, limits):
    eprint("Computing ISOHash for {}".format(path))
    context = get_context_by_suffix(path)
    if context == "wcnf":
        ihash = wcnfisohash(path)
    else:
        ihash = isohash(path)
    return [("isohash", hash, ihash)]


## Base Features
def compute_base_features(hash, path, limits, tp=None):
    eprint("Extracting base features from {} {}".format(hash, path))
    rec = extract_base_features(path, limits["tlim"], limits["mlim"])
    return [(key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items()]


## Gate Features
def compute_gate_features(hash, path, limits, tp=None):
    eprint("Extracting gate features from {} {}".format(hash, path))
    rec = extract_gate_features(path, limits["tlim"], limits["mlim"])
    return [(key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items()]


## WCNF Base Features
def compute_wcnf_base_features(hash, path, limits, tp=None):
    eprint("Extracting WCNF base features from {} {}".format(hash, path))
    rec = extract_wcnf_base_features(path, limits["tlim"], limits["mlim"])
    return [(key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items()]


## OPB Base Features
def compute_opb_base_features(hash, path, limits, tp=None):
    eprint("Extracting OPB base features from {} {}".format(hash, path))
    rec = extract_opb_base_features(path, limits["tlim"], limits["mlim"])
    return [(key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items()]


## SANI Features
def compute_sani_features(hash, path, limits, tp=None):
    eprint("Extracting SANI features from {} {}".format(hash, path))
    rec = checksani(path, limits["tlim"], limits["mlim"])
    return [(key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items()]


generic_extractors = {
    "base": {
        "description": "Extract base features from CNF files. ",
        "contexts": ["cnf", "sancnf"],
        "features": [(name, "empty") for name in base_feature_names()],
        "compute": compute_base_features,
    },
    "checksani": {
        "description": "Extract sanitise status from CNF files. ",
        "contexts": ["cnf", "sancnf"],
        "features": [(name, "empty") for name in checksani_feature_names()],
        "compute": compute_sani_features,
    },
    "gate": {
        "description": "Extract gate features from CNF files. ",
        "contexts": ["cnf", "sancnf"],
        "features": [(name, "empty") for name in gate_feature_names()],
        "compute": compute_gate_features,
    },
    "isohash": {
        "description": "Compute ISOHash for CNF or WCNF files. ",
        "contexts": ["cnf", "wcnf", "sancnf"],
        "features": [("isohash", "empty")],
        "compute": compute_isohash,
    },
    "wcnfbase": {
        "description": "Extract base features from WCNF files. ",
        "contexts": ["wcnf"],
        "features": [(name, "empty") for name in wcnf_base_feature_names()],
        "compute": compute_wcnf_base_features,
    },
    "opbbase": {
        "description": "Extract base features from OPB files. ",
        "contexts": ["opb"],
        "features": [(name, "empty") for name in opb_base_feature_names()],
        "compute": compute_opb_base_features,
    },
}


def init_features_generic(key: str, api: GBD, rlimits, df: pl.DataFrame, target_db):
    einfo = generic_extractors[key]
    context = api.database.dcontext(target_db)
    if not context in einfo["contexts"]:
        raise InitializerException("Target database context must be in {}".format(einfo["contexts"]))
    extractor = Initializer(api, rlimits, target_db, einfo["features"], einfo["compute"])
    extractor.create_features()
    extractor.run(df)


def init_local(api: GBD, rlimits, root, target_db):
    context = api.database.dcontext(target_db)

    features = [("local", None), ("filename", None)]
    extractor = Initializer(api, rlimits, target_db, features, compute_hash)
    extractor.create_features()

    # Cleanup stale entries
    df: pl.DataFrame = api.query(group_by=context + ":local", collapse=None)
     
    def path_exists(p):
        return p is not None and os.path.exists(p)
    
    missing = df.with_columns(
        exists=pl.col("local").map_elements(
            path_exists,
            return_dtype=pl.Boolean
        )
    ).filter(~pl.col("exists")).select("local")
    
    if len(missing) and api.verbose:
        for path in missing["local"].to_list():
            eprint(path)
    if len(missing) and confirm("{} files not found. Remove stale entries from local table?".format(len(missing))):
        api.reset_values("local", values=missing["local"].to_list())

    # Create df with paths not yet in local table
    paths = [path for suffix in suffixes(context) for path in glob.iglob(root + "/**/*" + suffix, recursive=True)]
    df2 = pl.DataFrame([(None, path) for path in paths if path not in df["local"].to_list()], schema=["hash", "local"], orient="row")

    extractor.run(df2)



# MIT License

# Copyright (c) 2025 Ashlin Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


from gbd_core.database import Database, DatabaseException
from gbd_core.grammar import Parser
from gbd_core import contexts
from gbd_core.schema import Schema


class GBDQuery:
    def __init__(self, db: Database, query):
        self.db = db
        self.parser = Parser(query)
        self.features = self.parser.get_features()

    def features_exist_or_throw(self, features):
        for feature in features:
            self.db.find(feature)

    # Generate SQL Query from given GBD Query
    def build_query(self, hashes=[], resolve=[], group_by=None, join_type="LEFT", collapse=None):
        group = group_by or self.determine_group_by(resolve)

        self.features_exist_or_throw(resolve + [group] + list(self.features))

        sql_select = self.build_select(group, resolve, collapse)

        sql_from = self.build_from(group, set(resolve) | self.features, join_type)

        sql_where = self.build_where(hashes, group)

        sql_groupby = "GROUP BY {}".format(self.db.faddr(group)) if collapse else ""
        sql_orderby = "ORDER BY {}".format(self.db.faddr(group))

        return "{} {} WHERE {} {} {}".format(sql_select, sql_from, sql_where, sql_groupby, sql_orderby)

    def determine_group_by(self, resolve):
        if len(resolve) == 0:
            return self.db.dcontext(self.db.find("hash").database) + ":hash"
        else:
            return self.db.dcontext(self.db.find(resolve[0]).database) + ":hash"

    def build_select(self, group_by, resolve, collapse=None):
        result = [self.db.faddr(f) for f in [group_by] + resolve]
        if collapse and collapse != "none":
            result = ["{}(DISTINCT {})".format(collapse, r) for r in result]
        return "SELECT DISTINCT " + ", ".join(result)

    def find_translator_feature(self, source_context, target_context):
        for dbname in self.db.get_databases(source_context):
            # eprint("Checking database {} for translator".format(dbname))
            if "to_" + target_context in self.db.get_features([dbname]):
                return self.db.find("to_" + target_context, dbname)

        for dbname in self.db.get_databases(target_context):
            # eprint("Checking database {} for translator".format(dbname))
            if "to_" + source_context in self.db.get_features([dbname]):
                return self.db.find("to_" + source_context, dbname)

        raise DatabaseException("No translator feature found for contexts {} and {}".format(source_context, target_context))

    def build_from(self, group, features, join_type="LEFT"):
        result = dict()

        gdatabase = self.db.find(group).database
        gtable = self.db.find(group).table
        gcontext = self.db.dcontext(gdatabase)
        gaddress = gdatabase + "." + gtable
        result[gaddress] = "FROM {}".format(gaddress)

        tables = set([(finfo.database, finfo.table) for finfo in [self.db.find(f) for f in features]])
        for fdatabase, ftable in tables:
            faddress = fdatabase + "." + ftable
            ffeatures_address = fdatabase + ".features"
            if not faddress in result:  # join only once
                fcontext = self.db.dcontext(fdatabase)
                if fcontext == gcontext:
                    if faddress == ffeatures_address:  # join features table directly
                        result[faddress] = "{j} JOIN {t} ON {t}.hash = {g}.hash".format(j=join_type, t=ffeatures_address, g=gaddress)
                    else:  # join non-unique features table via features table
                        fname = ftable
                        if not ffeatures_address in result:
                            result[ffeatures_address] = "{j} JOIN {t} ON {t}.hash = {g}.hash".format(j=join_type, t=ffeatures_address, g=gaddress)
                        result[faddress] = "{j} JOIN {t} ON {t}.hash = {ft}.{n}".format(j=join_type, t=faddress, ft=ffeatures_address, n=fname)
                else:
                    tfeat = self.find_translator_feature(gcontext, fcontext)
                    direction = ("hash", "value") if self.db.dcontext(tfeat.database) == gcontext else ("value", "hash")

                    taddress = tfeat.database + "." + tfeat.table
                    if not taddress in result:
                        result[taddress] = "INNER JOIN {trans} ON {group}.hash = {trans}.{dir0}".format(trans=taddress, group=gaddress, dir0=direction[0])

                    result[faddress] = "INNER JOIN {feat} ON {trans}.{dir1} = {feat}.hash".format(feat=faddress, trans=taddress, dir1=direction[1])

        return " ".join(result.values())

    def build_where(self, hashes, group_by):
        group_column = self.db.faddr(group_by)
        group_table = self.db.faddr_table(group_by)
        result = group_column + " != 'None' AND " + self.parser.get_sql(self.db)
        if len(hashes):
            result = result + " AND {}.hash in ('{}')".format(group_table, "', '".join(hashes))
        return result




import os
import unittest
import polars as pl
import random
import sqlite3

from gbd_core.database import Database
from gbd_core.schema import Schema
from gbd_core.api import GBD, GBDException
from gbd_init.initializer import Initializer
from gbd_core.contexts import identify
from gbd_init.feature_extractors import init_local, init_features_generic, generic_extractors

from tests import util

class InitTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.file = util.get_random_unique_filename('test', '.db')
        sqlite3.connect(self.file).close()
        self.name = Schema.dbname_from_path(self.file)
        self.db = Database([self.file], verbose=False)
        self.benchmark = "benchmark.cnf"
        self.dir = os.path.dirname(os.path.realpath(self.benchmark))
        with open(self.benchmark, 'w') as file:            
            file.write(util.get_random_formula(20))
        self.reference_hash = identify(self.benchmark)
        return super().setUp()

    def tearDown(self) -> None:
        if os.path.exists(self.file):
            os.remove(self.file)
        if os.path.exists(self.benchmark):
            os.remove(self.benchmark)
        return super().tearDown()

    def init_random(self, hash, path, limits):
        return [ ('random', hash, random.randint(1, 1000)) ]

    def test_init_random(self):
        api = GBD([self.file], verbose=False)
        rlimits = { 'jobs': 1, 'tlim': 5000, 'mlim': 2000, 'flim': 1000 }
        init = Initializer(api, rlimits, self.name, [('random', 0)], self.init_random)
        init.create_features()
        self.assertTrue(api.feature_exists('random'))
        df: pl.DataFrame  = pl.DataFrame([(str(n), None) for n in range(100)], schema=["hash", "local"], orient="row")
        init.run(df)
        df: pl.DataFrame = api.query("random > 0", [], ["random"])
        self.assertEqual(len(df), 100)

    def test_init_local(self):
        api = GBD([self.file], verbose=False)
        rlimits = { 'jobs': 1, 'tlim': 5000, 'mlim': 2000, 'flim': 1000 }
        init_local(api, rlimits, self.dir, self.name)
        self.assertTrue(api.feature_exists('local'))
        df: pl.DataFrame = api.query("local like %benchmark.cnf", [], ["local"])
        self.assertEqual(len(df), 1)
        self.assertEqual(df.to_dicts()[0]['local'], os.path.realpath(self.benchmark))
        self.assertEqual(df.to_dicts()[0]['hash'], self.reference_hash)

    def test_init_cnf_features_generic(self):
        api = GBD([self.file], verbose=False)
        rlimits = { 'jobs': 1, 'tlim': 5000, 'mlim': 2000, 'flim': 1000 }
        init_local(api, rlimits, self.dir, self.name)
        df: pl.DataFrame = api.query("local like %benchmark.cnf", [], ["local"])
        for key in generic_extractors.keys():
            if 'cnf' in generic_extractors[key]['contexts']:
                init_features_generic(key, api, rlimits, df, self.name)
                for feature in generic_extractors[key]['features']:
                    self.assertTrue(api.feature_exists(feature[0]))

    


# MIT License

# Copyright (c) 2025 Ashlin Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

from gbd_init.gbdhash import cnf_hash, opb_hash, wcnf_hash

### Default Context
default = "cnf"

### Configuration of Available Contexts
config = {
    "cnf": {
        "description": "Conjunctive Normal Form (CNF) in DIMACS format",
        "suffix": ".cnf",
        "idfunc": cnf_hash,
    },
    "sancnf": {
        "description": "Sanitized Conjunctive Normal Form (CNF) in DIMACS format",
        "suffix": ".sanitized.cnf",
        "idfunc": cnf_hash,
    },
    "kis": {
        "description": "k-Independent Set (KIS) in DIMACS-like graph format",
        "suffix": ".kis",
        "idfunc": cnf_hash,
    },
    "opb": {
        "description": "Pseudo-Boolean Optimization Problem in OPB format",
        "suffix": ".opb",
        "idfunc": opb_hash,
    },
    "wecnf": {
        "description": "Weighted Extended Conjunctive Normal Form (WECNF)",
        "suffix": ".wecnf",
        "idfunc": cnf_hash,
    },
    "wcnf": {
        "description": "MaxSAT instances in WCNF format",
        "suffix": ".wcnf",
        "idfunc": wcnf_hash,
    },
}


def description(context):
    return config[context]["description"]


def suffixes(context):
    return [config[context]["suffix"] + p for p in ["", ".gz", ".lzma", ".xz", ".bz2"]]


def idfunc(context):
    return config[context]["idfunc"]


def contexts():
    return config.keys()


def default_context():
    return default


def get_context_by_suffix(benchmark):
    for context in contexts():
        for suffix in suffixes(context):
            if benchmark.endswith(suffix):
                return context
    return None


def identify(path, ct=None):
    context = ct or get_context_by_suffix(path)
    if context is None:
        raise Exception("Unable to associate context: " + path)
    else:
        idf = idfunc(context)
        return idf(path)



# MIT License

# Copyright (c) 2025 Ashlin Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


import os
import polars as pl
from functools import reduce

from gbd_core import contexts
from gbd_core.api import GBD, GBDException
from gbd_core import util

from gbd_core.contexts import identify
from gbd_init.initializer import Initializer, InitializerException

try:
    from gbdc import cnf2kis, sanitise, normalise
except ImportError:

    def cnf2kis(ipath, opath):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def sanitise(ipath, opath):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def normalise(ipath, opath):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")


def kis_filename(path):
    kispath = reduce(lambda path, suffix: path[: -len(suffix)] if path.endswith(suffix) else path, contexts.suffixes("cnf"), path)
    return kispath + ".kis"


def sanitised_filename(path):
    sanpath = reduce(lambda path, suffix: path[: -len(suffix)] if path.endswith(suffix) else path, contexts.suffixes("cnf"), path)
    return sanpath + ".sanitized.cnf"


def normalised_filename(path):
    normpath = reduce(lambda path, suffix: path[: -len(suffix)] if path.endswith(suffix) else path, contexts.suffixes("cnf"), path)
    return normpath + ".normalised.cnf"


def wrap_cnf2kis(hash, path, limits):
    kispath = kis_filename(path)
    util.eprint("Transforming {} to k-ISP {}".format(path, kispath))
    try:
        result = cnf2kis(path, kispath)
        if "local" in result:
            kishash = result["hash"]
            return [
                ("local", kishash, result["local"]),
                ("to_cnf", kishash, hash),
                ("nodes", kishash, result["nodes"]),
                ("edges", kishash, result["edges"]),
                ("k", kishash, result["k"]),
            ]
        else:
            raise GBDException("CNF2KIS failed for {} due to {}".format(path, result["hash"]))
    except Exception as e:
        util.eprint(str(e))
        if os.path.exists(kispath):
            os.remove(kispath)

    return []


def wrap_sanitise(hash, path, limits):
    sanpath = sanitised_filename(path)
    util.eprint("Sanitising {}".format(path))
    try:
        with open(sanpath, "w") as f, util.stdout_redirected(f):
            result = sanitise(path, sanpath)
            if "local" in result:
                sanhash = result["hash"]
                return [("local", sanhash, result["local"]), ("to_cnf", sanhash, hash)]
            else:
                raise GBDException("Sanitization failed for {}".format(path))
    except Exception as e:
        util.eprint(str(e))
        if os.path.exists(sanpath):
            os.remove(sanpath)

    return []


def wrap_normalise(hash, path, limits):
    normpath = normalised_filename(path)
    util.eprint("Normalising {}".format(path))
    try:
        with open(normpath, "w") as f, util.stdout_redirected(f):
            result = normalise(path, normpath)
            normhash = result["hash"]
            if "local" in result and hash == normhash:
                return [("local", normhash, result["local"])]
            else:
                raise GBDException("Normalisation failed for {}".format(path))
    except Exception as e:
        util.eprint(str(e))
        if os.path.exists(normpath):
            os.remove(normpath)

    return []


def transform_instances_generic(key: str, api: GBD, rlimits, query, hashes, target_db, source, collapse=None):
    einfo = generic_transformers[key]
    context = api.database.dcontext(target_db)
    if not context in einfo["target"]:
        raise InitializerException("Target database context must be in {}".format(einfo["target"]))
    if not source in einfo["source"]:
        raise InitializerException("Source database context must be in {}".format(einfo["source"]))
    transformer = Initializer(api, rlimits, target_db, einfo["features"], einfo["compute"])
    transformer.create_features()
    
    def path_exists(p):
        return p is not None and os.path.exists(einfo["filename"](p))

    df: pl.DataFrame = api.query(query, hashes, [source + ":local"], collapse=collapse)
    missing = df.with_columns(
        exists=pl.col("local").map_elements(
            path_exists,
            return_dtype=pl.Boolean
        )
    ).filter(~pl.col("exists"))

    transformer.run(missing)


generic_transformers = {
    "sanitise": {
        "description": "Sanitise CNF files. ",
        "source": ["cnf"],
        "target": ["sancnf"],
        "features": [("local", None), ("to_cnf", None)],
        "compute": wrap_sanitise,
        "filename": sanitised_filename,
    },
    "normalise": {
        "description": "Normalise CNF files. ",
        "source": ["cnf"],
        "target": ["cnf"],
        "features": [("local", None)],
        "compute": wrap_normalise,
        "filename": normalised_filename,
    },
    "cnf2kis": {
        "description": "Transform CNF files to k-ISP instances. ",
        "source": ["cnf"],
        "target": ["kis"],
        "features": [("local", None), ("to_cnf", None), ("nodes", "empty"), ("edges", "empty"), ("k", "empty")],
        "compute": wrap_cnf2kis,
        "filename": kis_filename,
    },
}



##!/usr/bin/python3
# -*- coding: utf-8 -*-

import os
import unittest
import sqlite3

from gbd_core.database import Database
from gbd_core.schema import Schema

from tests import util

class SchemaTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.file = util.get_random_unique_filename('test', '.db')
        sqlite3.connect(self.file).close()
        self.name = Schema.dbname_from_path(self.file)
        self.db = Database([self.file], verbose=False)
        return super().setUp()

    def tearDown(self) -> None:
        if os.path.exists(self.file):
            os.remove(self.file)
        return super().tearDown()

    def test_create_db(self):
        self.assertTrue(Schema.is_database(self.file))
        self.assertEqual(len(self.db.get_databases()), 1)
        self.assertTrue(self.db.dexists(self.name))
        
    def test_create_unique_feature(self):
        FEAT = "featA"
        self.db.create_feature(FEAT, default_value="empty")
        self.assertIn(FEAT, self.db.get_features())
        self.assertIn("features", self.db.get_tables())
        finfo = self.db.find(FEAT)
        self.assertEqual(finfo.table, "features")
        self.assertEqual(finfo.column, FEAT)
        self.assertEqual(finfo.default, "empty")
        self.assertEqual(finfo.database, self.name)
        
    def test_create_nonunique_feature(self):
        FEAT = "featB"
        self.db.create_feature(FEAT, default_value=None)
        self.assertIn(FEAT, self.db.get_features())
        self.assertIn("features", self.db.get_tables())
        finfo = self.db.find(FEAT)
        self.assertEqual(finfo.table, FEAT)
        self.assertEqual(finfo.column, "value")
        self.assertEqual(finfo.default, None)
        self.assertEqual(finfo.database, self.name)




##!/usr/bin/python3
# -*- coding: utf-8 -*-

import os
import unittest
import sqlite3

from gbd_core.query import GBDQuery
from gbd_core.database import Database, DatabaseException
from gbd_core.schema import Schema
from tests import util

class DatabaseTestCase(unittest.TestCase):

    feat = "nonunique_feature"
    val1 = "value1"
    val2 = "value2"

    def setUp(self) -> None:
        self.file = util.get_random_unique_filename('test', '.db')
        sqlite3.connect(self.file).close()
        self.name = Schema.dbname_from_path(self.file)
        self.db = Database([self.file], verbose=False)
        self.db.create_feature(self.feat, default_value=None)
        self.db.set_values(self.feat, self.val1, ["a", "b", "c"])
        self.db.set_values(self.feat, self.val2, ["a", "b", "c"])
        return super().setUp()

    def tearDown(self) -> None:
        if os.path.exists(self.file):
            os.remove(self.file)
        return super().tearDown()

    def query(self, feat, val):
        qb = GBDQuery(self.db, "{}={}".format(feat, val))
        q = qb.build_query()
        return [ hash for (hash, ) in self.db.query(q) ]

    def dump(self):
        import sqlite3
        conn = sqlite3.connect(self.file)
        for line in conn.iterdump():
            print(line)
        conn.close()

    # Test that the feature values are initialized correctly in test setup
    def test_feature_values_exist(self):
        res = self.query(self.feat, self.val1)
        self.assertEqual(len(res), 3)
        self.assertSetEqual(set(res), set(["a", "b", "c"]))
        res = self.query(self.feat, self.val2)
        self.assertEqual(len(res), 3)
        self.assertSetEqual(set(res), set(["a", "b", "c"]))

    # Delete specific hash-value pair and check that it is gone and the others are still there
    def test_feature_values_delete_hash_value(self):
        self.db.delete(self.feat, [ self.val1 ], ["a"])
        res = self.query(self.feat, self.val1)
        self.assertEqual(len(res), 2)
        self.assertSetEqual(set(res), set(["b", "c"]))
        res = self.query(self.feat, self.val2)
        self.assertEqual(len(res), 3)
        self.assertSetEqual(set(res), set(["a", "b", "c"]))

    # Delete specific hash and check that it is gone and the others are still there
    def test_feature_values_delete_hash(self):
        self.db.delete(self.feat, [ ], ["a"])
        res = self.query(self.feat, self.val1)
        self.assertEqual(len(res), 2)
        self.assertSetEqual(set(res), set(["b", "c"]))
        res = self.query(self.feat, self.val2)
        self.assertEqual(len(res), 2)
        self.assertSetEqual(set(res), set(["b", "c"]))
        res = self.query(self.feat, "None")
        self.assertEqual(len(res), 1)
        self.assertSetEqual(set(res), set(["a"]))

    # Delete specific value and check that it is gone and the others are still there
    def test_feature_values_delete_value(self):
        self.db.delete(self.feat, [ self.val1 ], [ ])
        res = self.query(self.feat, self.val1)
        self.assertEqual(len(res), 0)
        res = self.query(self.feat, self.val2)
        self.assertEqual(len(res), 3)
        self.assertSetEqual(set(res), set([ "a", "b", "c" ]))

    # Delete feature
    def test_nonunique_feature_delete(self):
        self.db.delete_feature(self.feat)
        self.assertRaises(DatabaseException, self.db.find, self.feat)